param ([string] $NavIde)

# If $NavIde is not provided try to find finsql in:
#    1) the current folder, or
#    2) RTC's installation folder.
if (-not $NavIde)
{
    if (Test-Path (Join-Path $PSScriptRoot finsql.exe))
    {
        $NavIde = (Join-Path $PSScriptRoot finsql.exe)
    }
    else
    {
        if ([Environment]::Is64BitProcess)
        {
            $RtcKey = 'HKLM:\SOFTWARE\Wow6432Node\Microsoft\Microsoft Dynamics NAV\110\RoleTailored Client'
        }
        else
        {
            $RtcKey = 'HKLM:\SOFTWARE\Microsoft\Microsoft Dynamics NAV\110\RoleTailored Client'
        }

        if (Test-Path $RtcKey)
        {
            $NavIde = (Join-Path (Get-ItemProperty $RtcKey).Path finsql.exe)
        }
    }
}

if ($NavIde -and (Test-Path $NavIde))
{
	$NavClientPath = (Get-Item $NavIde).DirectoryName
}

<#
    .SYNOPSIS
    Imports NAV application objects from a file into a database.

    .DESCRIPTION
    The Import-NAVApplicationObject function imports the objects from the specified file(s) into the specified database. When multiple files are specified, finsql is invoked for each file. For better performance the files can be joined first. However, using seperate files can be useful for analyzing import errors.

    .INPUTS
    System.String[]
    You can pipe a path to the Import-NavApplicationObject function.

    .OUTPUTS
    None

    .EXAMPLE
    Import-NAVApplicationObject MyAppSrc.txt MyApp
    This command imports all application objects in MyAppSrc.txt into the MyApp database.

    .EXAMPLE
    Import-NAVApplicationObject MyAppSrc.txt -DatabaseName MyApp
    This command imports all application objects in MyAppSrc.txt into the MyApp database.

    .EXAMPLE
    Get-ChildItem MyAppSrc | Import-NAVApplicationObject -DatabaseName MyApp
    This commands imports all objects in all files in the MyAppSrc folder into the MyApp database. The files are imported one by one.

    .EXAMPLE
    Get-ChildItem MyAppSrc | Join-NAVApplicationObject -Destination .\MyAppSrc.txt -PassThru | Import-NAVApplicationObject -Database MyApp
    This commands joins all objects in all files in the MyAppSrc folder into a single file and then imports them in the MyApp database.
#>
function Import-NAVApplicationObject
{
    [CmdletBinding(DefaultParameterSetName="All", SupportsShouldProcess=$true, ConfirmImpact='High')]
    Param(
        # Specifies one or more files to import.
        [Parameter(Mandatory=$true, Position=0, ValueFromPipeline=$true, ValueFromPipelineByPropertyName=$true)]
        [Alias('PSPath')]
        [string[]] $Path,

        # Specifies the name of the database into which you want to import.
        [Parameter(Mandatory=$true, Position=1)]
        [string] $DatabaseName,

        # Specifies the name of the SQL server instance to which the database you want to import into is attached. The default value is the default instance on the local host (.).
        [ValidateNotNullOrEmpty()]
        [string] $DatabaseServer = '.',

        # Specifies the log folder.
        [ValidateNotNullOrEmpty()]
        [string] $LogPath = "$Env:TEMP\NavIde\$([GUID]::NewGuid().GUID)",

        # Specifies the import action. The default value is 'Default'.
        [ValidateSet('Default','Overwrite','Skip')] [string] $ImportAction = 'Default',

        # Specifies the schema synchronization behaviour. The default value is 'Yes'.
        [ValidateSet('Yes','No','Force')] [string] $SynchronizeSchemaChanges = 'Yes',

        # The user name to use to authenticate to the database. The user name must exist in the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Username,

        # The password to use with the username parameter to authenticate to the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Password,

        # Specifies the name of the server that hosts the Microsoft Dynamics NAV Server instance, such as MyServer.
        [ValidateNotNullOrEmpty()]
        [string] $NavServerName,

        # Specifies the Microsoft Dynamics NAV Server instance that is being used.The default value is DynamicsNAV90.
        [ValidateNotNullOrEmpty()]
        [string] $NavServerInstance = "DynamicsNAV90",

        # Specifies the port on the Microsoft Dynamics NAV Server server that the Microsoft Dynamics NAV Windows PowerShell cmdlets access. The default value is 7045.
        [ValidateNotNullOrEmpty()]
        [int16]  $NavServerManagementPort = 7045)

    PROCESS
    {
        if ($Path.Count -eq 1)
        {
            $Path = (Get-Item $Path).FullName
        }

        if ($PSCmdlet.ShouldProcess(
            "Import application objects from $Path into the $DatabaseName database.",
            "Import application objects from $Path into the $DatabaseName database. If you continue, you may lose data in fields that are removed or changed in the imported file.",
            'Confirm'))
        {
            $navServerInfo = GetNavServerInfo $NavServerName $NavServerInstance $NavServerManagementPort

            foreach ($file in $Path)
            {
                # Log file name is based on the name of the imported file.
                $logFile = "$LogPath\$((Get-Item $file).BaseName).log"
                $command = "Command=ImportObjects`,ImportAction=$ImportAction`,SynchronizeSchemaChanges=$SynchronizeSchemaChanges`,File=`"$file`""

                try
                {
                    RunNavIdeCommand -Command $command `
                                     -DatabaseServer $DatabaseServer `
                                     -DatabaseName $DatabaseName `
                                     -NTAuthentication:($Username -eq $null) `
                                     -Username $Username `
                                     -Password $Password `
                                     -NavServerInfo $navServerInfo `
                                     -LogFile $logFile `
                                     -ErrText "Error while importing $file" `
                                     -Verbose:$VerbosePreference
                }
                catch
                {
                    Write-Error $_
                }
            }
        }
    }
}

function GetNavServerInfo
(
    [string] $NavServerName,
    [string] $NavServerInstance,
    [int16]  $NavServerManagementPort
)
{
    $navServerInfo = ""
    if ($navServerName)
    {
        $navServerInfo = @"
`,NavServerName="$NavServerName"`,NavServerInstance="$NavServerInstance"`,NavServerManagementport=$NavServerManagementPort
"@
    }

    $navServerInfo
}

function RunNavIdeCommand
{
    [CmdletBinding()]
    Param(
    [string] $Command,
    [string] $DatabaseServer,
    [string] $DatabaseName,
    [switch] $NTAuthentication,
    [string] $Username,
    [string] $Password,
    [string] $NavServerInfo,
    [string] $LogFile,
    [string] $ErrText)

    TestNavIde
    $logPath = (Split-Path $LogFile)

    Remove-Item "$logPath\navcommandresult.txt" -ErrorAction Ignore
    Remove-Item $logFile -ErrorAction Ignore

    $databaseInfo = @"
ServerName="$DatabaseServer"`,Database="$DatabaseName"
"@
    if ($Username)
    {
        $databaseInfo = @"
ntauthentication=No`,username="$Username"`,password="$Password"`,$databaseInfo
"@
    }

    $finSqlCommand = @"
& "$NavIde" --% $Command`,LogFile="$logFile"`,${databaseInfo}${NavServerInfo} | Out-Null
"@

    Write-Verbose "Running command: $finSqlCommand"
    Invoke-Expression -Command  $finSqlCommand

    if (Test-Path "$logPath\navcommandresult.txt")
    {
        if (Test-Path $LogFile)
        {
            throw "${ErrorText}: $(Get-Content $LogFile -Raw)" -replace "`r[^`n]","`r`n"
        }
    }
    else
    {
        throw "${ErrorText}!"
    }
}

<#
    .SYNOPSIS
    Export NAV application objects from a database into a file.

    .DESCRIPTION
    The Export-NAVApplicationObject function exports the objects from the specified database into the specified file. A filter can be specified to select the application objects to be exported.

    .INPUTS
    None
    You cannot pipe input to this function.

    .OUTPUTS
    System.IO.FileInfo
    An object representing the exported file.

    .EXAMPLE
    Export-NAVApplicationObject MyApp MyAppSrc.txt
    Exports all application objects from the MyApp database to MyAppSrc.txt.

    .EXAMPLE
    Export-NAVApplicationObject MyAppSrc.txt -DatabaseName MyApp
    Exports all application objects from the MyApp database to MyAppSrc.txt.

    .EXAMPLE
    Export-NAVApplicationObject MyApp COD1-10.txt -Filter 'Type=Codeunit;Id=1..10'
    Exports codeunits 1..10 from the MyApp database to COD1-10.txt

    .EXAMPLE
    Export-NAVApplicationObject COD1-10.txt -DatabaseName MyApp -Filter 'Type=Codeunit;Id=1..10'
    Exports codeunits 1..10 from the MyApp database to COD1-10.txt

    .EXAMPLE
    Export-NAVApplicationObject COD1-10.txt -DatabaseName MyApp -Filter 'Type=Codeunit;Id=1..10' | Import-NAVApplicationObject -DatabaseName MyApp2
    Copies codeunits 1..10 from the MyApp database to the MyApp2 database.

    .EXAMPLE
    Export-NAVApplicationObject MyAppSrc.txt -DatabaseName MyApp | Split-NAVApplicationObject -Destination MyAppSrc
    Exports all application objects from the MyApp database and splits into single-object files in the MyAppSrc folder.
#>
function Export-NAVApplicationObject
{
    [CmdletBinding(DefaultParameterSetName="All",SupportsShouldProcess = $true)]
    Param(
        # Specifies the name of the database from which you want to export.
        [Parameter(Mandatory=$true, Position=0)]
        [string] $DatabaseName,

        # Specifies the file to export to.
        [Parameter(Mandatory=$true, Position=1)]
        [string] $Path,

        # Specifies the name of the SQL server instance to which the database you want to import into is attached. The default value is the default instance on the local host (.).
        [ValidateNotNullOrEmpty()]
        [string] $DatabaseServer = '.',

        # Specifies the log folder.
        [ValidateNotNullOrEmpty()]
        [string] $LogPath = "$Env:TEMP\NavIde\$([GUID]::NewGuid().GUID)",

        # Specifies the filter that selects the objects to export.
        [string] $Filter,

        # Allows the command to create a file that overwrites an existing file.
        [Switch] $Force,

        # Allows the command to skip application objects that are excluded from license, when exporting as txt.
        [Switch] $ExportTxtSkipUnlicensed,

        # Export the application object to the syntax supported by the Txt2Al converter.
        [Switch] $ExportToNewSyntax,

        # The user name to use to authenticate to the database. The user name must exist in the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Username,

        # The password to use with the username parameter to authenticate to the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Password)

    if ($PSCmdlet.ShouldProcess(
        "Export application objects from $DatabaseName database to $Path.",
        "Export application objects from $DatabaseName database to $Path.",
        'Confirm'))
    {
        if (!$Force -and (Test-Path $Path) -and !$PSCmdlet.ShouldContinue(
            "$Path already exists. If you continue, $Path will be overwritten.",
            'Confirm'))
        {
            Write-Error "$Path already exists."
            return
        }
    }
    else
    {
        return
    }

    $skipUnlicensed = "0"
    if($ExportTxtSkipUnlicensed)
    {
        $skipUnlicensed = "1"
    }

    $exportCommand = "ExportObjects"
    if($ExportToNewSyntax)
    {
        $exportCommand = "ExportToNewSyntax"
    }

    $command = "Command=$exportCommand`,ExportTxtSkipUnlicensed=$skipUnlicensed`,File=`"$Path`""
    if($Filter)
    {
        $command = "$command`,Filter=`"$Filter`""
    }

    $logFile = (Join-Path $logPath naverrorlog.txt)

    try
    {
        RunNavIdeCommand -Command $command `
                         -DatabaseServer $DatabaseServer `
                         -DatabaseName $DatabaseName `
                         -NTAuthentication:($Username -eq $null) `
                         -Username $Username `
                         -Password $Password `
                         -NavServerInfo "" `
                         -LogFile $logFile `
                         -ErrText "Error while exporting $Filter" `
                         -Verbose:$VerbosePreference
        Get-Item $Path
    }
    catch
    {
        Write-Error $_
    }
}

<#
    .SYNOPSIS
    Deletes NAV application objects from a database.

    .DESCRIPTION
    The Delete-NAVApplicationObject function deletes objects from the specified database. A filter can be specified to select the application objects to be deleted.

    .INPUTS
    None
    You cannot pipe input to this function.

    .OUTPUTS
    None

    .EXAMPLE
    Delete-NAVApplicationObject -DatabaseName MyApp -Filter 'Type=Codeunit;Id=1..10'
    Deletes codeunits 1..10 from the MyApp database
#>
function Delete-NAVApplicationObject
{
    [CmdletBinding(DefaultParameterSetName="All", SupportsShouldProcess=$true, ConfirmImpact='High')]
    Param(
        # Specifies the name of the database from which you want to delete objects.
        [Parameter(Mandatory=$true, Position=0)]
        [string] $DatabaseName,

        # Specifies the name of the SQL server instance to which the database you want to delete objects from is attached. The default value is the default instance on the local host (.).
        [ValidateNotNullOrEmpty()]
        [string] $DatabaseServer = '.',

        # Specifies the log folder.
        [ValidateNotNullOrEmpty()]
        [string] $LogPath = "$Env:TEMP\NavIde\$([GUID]::NewGuid().GUID)",

        # Specifies the filter that selects the objects to delete.
        [string] $Filter,

        # Specifies the schema synchronization behaviour. The default value is 'Yes'.
        [ValidateSet('Yes','No','Force')]
        [string] $SynchronizeSchemaChanges = 'Yes',

        # The user name to use to authenticate to the database. The user name must exist in the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Username,

        # The password to use with the username parameter to authenticate to the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Password,

        # Specifies the name of the server that hosts the Microsoft Dynamics NAV Server instance, such as MyServer.
        [ValidateNotNullOrEmpty()]
        [string] $NavServerName,

        # Specifies the Microsoft Dynamics NAV Server instance that is being used.The default value is DynamicsNAV90.
        [ValidateNotNullOrEmpty()]
        [string] $NavServerInstance = "DynamicsNAV90",

        # Specifies the port on the Microsoft Dynamics NAV Server server that the Microsoft Dynamics NAV Windows PowerShell cmdlets access. The default value is 7045.
        [ValidateNotNullOrEmpty()]
        [int16]  $NavServerManagementPort = 7045)

    if ($PSCmdlet.ShouldProcess(
        "Delete application objects from $DatabaseName database.",
        "Delete application objects from $DatabaseName database.",
        'Confirm'))
    {
        $command = "Command=DeleteObjects`,SynchronizeSchemaChanges=$SynchronizeSchemaChanges"
        if($Filter)
        {
            $command = "$command`,Filter=`"$Filter`""
        }

        $logFile = (Join-Path $logPath naverrorlog.txt)
        $navServerInfo = GetNavServerInfo $NavServerName $NavServerInstance $NavServerManagementPort

        try
        {
            RunNavIdeCommand -Command $command `
                             -DatabaseServer $DatabaseServer `
                             -DatabaseName $DatabaseName `
                             -NTAuthentication:($Username -eq $null) `
                             -Username $Username `
                             -Password $Password `
                             -NavServerInfo $navServerInfo `
                             -LogFile $logFile `
                             -ErrText "Error while deleting $Filter" `
                             -Verbose:$VerbosePreference
        }
        catch
        {
            Write-Error $_
        }
    }
}

<#
    .SYNOPSIS
    Compiles NAV application objects in a database.

    .DESCRIPTION
    The Compile-NAVApplicationObject function compiles application objects in the specified database. A filter can be specified to select the application objects to be compiled. Unless the Recompile switch is used only uncompiled objects are compiled.

    .INPUTS
    None
    You cannot pipe input to this function.

    .OUTPUTS
    None

    .EXAMPLE
    Compile-NAVApplicationObject MyApp
    Compiles all uncompiled application objects in the MyApp database.

    .EXAMPLE
    Compile-NAVApplicationObject MyApp -Filter 'Type=Codeunit' -Recompile
    Compiles all codeunits in the MyApp database.

    .EXAMPLE
    'Page','Codeunit','Table','XMLport','Report' | % { Compile-NAVApplicationObject -Database MyApp -Filter "Type=$_" -AsJob } | Receive-Job -Wait
    Compiles all uncompiled Pages, Codeunits, Tables, XMLports, and Reports in the MyApp database in parallel and wait until it is done. Note that some objects may remain uncompiled due to race conditions. Those remaining objects can be compiled in a seperate command.

#>
function Compile-NAVApplicationObject
{
    [CmdletBinding(DefaultParameterSetName="All")]
    Param(
        # Specifies the name of the Dynamics NAV database.
        [Parameter(Mandatory=$true, Position=0)]
        [string] $DatabaseName,

        # Specifies the name of the SQL server instance to which the Dynamics NAV database is attached. The default value is the default instance on the local host (.).
        [ValidateNotNullOrEmpty()]
        [string] $DatabaseServer = '.',

        # Specifies the log folder.
        [ValidateNotNullOrEmpty()]
        [string] $LogPath = "$Env:TEMP\NavIde\$([GUID]::NewGuid().GUID)",

        # Specifies the filter that selects the objects to compile.
        [string] $Filter,

        # Compiles objects that are already compiled.
        [Switch] $Recompile,

        # Compiles in the background returning an object that represents the background job.
        [Switch] $AsJob,

        # Specifies the schema synchronization behaviour. The default value is 'Yes'.
        [ValidateSet('Yes','No','Force')]
        [string] $SynchronizeSchemaChanges = 'Yes',

        # The user name to use to authenticate to the database. The user name must exist in the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Username,

        # The password to use with the username parameter to authenticate to the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Password,

        # Specifies the name of the server that hosts the Microsoft Dynamics NAV Server instance, such as MyServer.
        [ValidateNotNullOrEmpty()]
        [string] $NavServerName,

        # Specifies the Microsoft Dynamics NAV Server instance that is being used.The default value is DynamicsNAV90.
        [ValidateNotNullOrEmpty()]
        [string] $NavServerInstance = "DynamicsNAV90",

        # Specifies the port on the Microsoft Dynamics NAV Server server that the Microsoft Dynamics NAV Windows PowerShell cmdlets access. The default value is 7045.
        [ValidateNotNullOrEmpty()]
        [int16]  $NavServerManagementPort = 7045)

    if (-not $Recompile)
    {
        $Filter += ';Compiled=0'
        $Filter = $Filter.TrimStart(';')
    }

    if ($AsJob)
    {
        $LogPath = "$LogPath\$([GUID]::NewGuid().GUID)"
        Remove-Item $LogPath -ErrorAction Ignore -Recurse -Confirm:$False -Force
        $scriptBlock =
        {
            Param($ScriptPath,$NavIde,$DatabaseName,$DatabaseServer,$LogPath,$Filter,$Recompile,$SynchronizeSchemaChanges,$Username,$Password,$NavServerName,$NavServerInstance,$NavServerManagementPort,$VerbosePreference)

            Import-Module "$ScriptPath\Microsoft.Dynamics.Nav.Ide.psm1" -ArgumentList $NavIde -Force -DisableNameChecking

            $args = @{
                DatabaseName = $DatabaseName
                DatabaseServer = $DatabaseServer
                LogPath = $LogPath
                Filter = $Filter
                Recompile = $Recompile
                SynchronizeSchemaChanges = $SynchronizeSchemaChanges
            }

            if($Username)
            {
                $args.Add("Username",$Username)
                $args.Add("Password",$Password)
            }

            if($NavServerName)
            {
                $args.Add("NavServerName",$NavServerName)
                $args.Add("NavServerInstance",$NavServerInstance)
                $args.Add("NavServerManagementPort",$NavServerManagementPort)
            }

            Compile-NAVApplicationObject @args -Verbose:$VerbosePreference
        }

        $job = Start-Job $scriptBlock -ArgumentList $PSScriptRoot,$NavIde,$DatabaseName,$DatabaseServer,$LogPath,$Filter,$Recompile,$SynchronizeSchemaChanges,$Username,$Password,$NavServerName,$NavServerInstance,$NavServerManagementPort,$VerbosePreference
        return $job
    }
    else
    {
        try
        {
            $logFile = (Join-Path $LogPath naverrorlog.txt)
            $navServerInfo = GetNavServerInfo $NavServerName $NavServerInstance $NavServerManagementPort
            $command = "Command=CompileObjects`,SynchronizeSchemaChanges=$SynchronizeSchemaChanges"
            if($Filter)
            {
                $command = "$command,Filter=`"$Filter`""
            }

            RunNavIdeCommand -Command $command `
                             -DatabaseServer $DatabaseServer `
                             -DatabaseName $DatabaseName `
                             -NTAuthentication:($Username -eq $null) `
                             -Username $Username `
                             -Password $Password `
                             -NavServerInfo $navServerInfo `
                             -LogFile $logFile `
                             -ErrText "Error while compiling $Filter" `
                             -Verbose:$VerbosePreference
        }
        catch
        {
            Write-Error $_
        }
    }
}

<#
    .SYNOPSIS
    Creates a new NAV application database.

    .DESCRIPTION
    The Create-NAVDatabase creates a new NAV database that includes the NAV system tables.

    .INPUTS
    None
    You cannot pipe input into this function.

    .OUTPUTS
    None

    .EXAMPLE
    Create-NAVDatabase MyNewApp
    Creates a new NAV database named MyNewApp.

    .EXAMPLE
    Create-NAVDatabase MyNewApp -ServerName "TestComputer01\NAVDEMO" -Collation "da-dk"
    Creates a new NAV database named MyNewApp on TestComputer01\NAVDEMO Sql server with Danish collation.
#>
function Create-NAVDatabase
{
    [CmdletBinding(DefaultParameterSetName="All")]
    Param(
         # Specifies the name of the Dynamics NAV database that will be created.
        [Parameter(Mandatory=$true, Position=0)]
        [string] $DatabaseName,

        # Specifies the name of the SQL server instance on which you want to create the database. The default value is the default instance on the local host (.).
        [ValidateNotNullOrEmpty()]
        [string] $DatabaseServer = '.',

        # Specifies the collation of the database.
        [ValidateNotNullOrEmpty()]
        [string] $Collation,

        # Specifies the log folder.
        [ValidateNotNullOrEmpty()]
        [string] $LogPath = "$Env:TEMP\NavIde\$([GUID]::NewGuid().GUID)",


        # The user name to use to authenticate to the database. The user name must exist in the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Username,

        # The password to use with the username parameter to authenticate to the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Password)

    $logFile = (Join-Path $LogPath naverrorlog.txt)

    $command = "Command=CreateDatabase`,Collation=$Collation"

    try
    {
        RunNavIdeCommand -Command $command `
                         -DatabaseServer $DatabaseServer `
                         -DatabaseName $DatabaseName `
                         -NTAuthentication:($Username -eq $null) `
                         -Username $Username `
                         -Password $Password `
                         -NavServerInfo $navServerInfo `
                         -LogFile $logFile `
                         -ErrText "Error while creating $DatabaseName" `
                         -Verbose:$VerbosePreference
    }
    catch
    {
        Write-Error $_
    }
}

<#
    .SYNOPSIS
    Performs a technical upgrade of a database from a previous version of Microsoft Dynamics NAV.

    .DESCRIPTION
    Performs a technical upgrade of a database from a previous version of Microsoft Dynamics NAV.

    .INPUTS
    None
    You cannot pipe input into this function.

    .OUTPUTS
    None

    .EXAMPLE
    Invoke-NAVDatabaseConversion MyApp
    Perform the technical upgrade on a NAV database named MyApp.

    .EXAMPLE
    Invoke-NAVDatabaseConversion MyApp -ServerName "TestComputer01\NAVDEMO"
    Perform the technical upgrade on a NAV database named MyApp on TestComputer01\NAVDEMO Sql server .
#>
function Invoke-NAVDatabaseConversion
{
    [CmdletBinding(DefaultParameterSetName="All")]
    Param(
         # Specifies the name of the Dynamics NAV database that will be created.
        [Parameter(Mandatory=$true, Position=0)]
        [string] $DatabaseName,

        # Specifies the name of the SQL server instance on which you want to create the database. The default value is the default instance on the local host (.).
        [ValidateNotNullOrEmpty()]
        [string] $DatabaseServer = '.',

        # Specifies the log folder.
        [ValidateNotNullOrEmpty()]
        [string] $LogPath = "$Env:TEMP\NavIde\$([GUID]::NewGuid().GUID)",

        # The user name to use to authenticate to the database. The user name must exist in the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Username,

        # The password to use with the username parameter to authenticate to the database. If you do not specify a user name and password, then the command uses the credentials of the current Windows user to authenticate to the database.
        [Parameter(Mandatory=$true, ParameterSetName="DatabaseAuthentication")]
        [string] $Password)

    $logFile = (Join-Path $LogPath naverrorlog.txt)

    $command = "Command=UpgradeDatabase"

    try
    {
        RunNavIdeCommand -Command $command `
                         -DatabaseServer $DatabaseServer `
                         -DatabaseName $DatabaseName `
                         -NTAuthentication:($Username -eq $null) `
                         -Username $Username `
                         -Password $Password `
                         -NavServerInfo "" `
                         -LogFile $logFile `
                         -ErrText "Error while converting $DatabaseName" `
                         -Verbose:$VerbosePreference
    }
    catch
    {
        Write-Error $_
    }
}

function TestNavIde
{
    if (-not $NavIde -or (($NavIde) -and -not (Test-Path $NavIde)))
    {
        throw '$NavIde was not correctly set. Please assign the path to finsql.exe to $NavIde ($NavIde = path).'
    }
}

Export-ModuleMember -Function *-* -Variable Nav*
# SIG # Begin signature block
# MIIkBQYJKoZIhvcNAQcCoIIj9jCCI/ICAQExDzANBglghkgBZQMEAgEFADB5Bgor
# BgEEAYI3AgEEoGswaTA0BgorBgEEAYI3AgEeMCYCAwEAAAQQH8w7YFlLCE63JNLG
# KX7zUQIBAAIBAAIBAAIBAAIBADAxMA0GCWCGSAFlAwQCAQUABCAMyCO7TOrRgPE8
# nTVKlWWyFTdMC2X4efB9gpP+QCuDg6CCDYIwggYAMIID6KADAgECAhMzAAAAww6b
# p9iy3PcsAAAAAADDMA0GCSqGSIb3DQEBCwUAMH4xCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jvc29mdCBDb2RlIFNpZ25p
# bmcgUENBIDIwMTEwHhcNMTcwODExMjAyMDI0WhcNMTgwODExMjAyMDI0WjB0MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMR4wHAYDVQQDExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24wggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIB
# AQC7V9c40bEGf0ktqW2zY596urY6IVu0mK6N1KSBoMV1xSzvgkAqt4FTd/NjAQq8
# zjeEA0BDV4JLzu0ftv2AbcnCkV0Fx9xWWQDhDOtX3v3xuJAnv3VK/HWycli2xUib
# M2IF0ZWUpb85Iq2NEk1GYtoyGc6qIlxWSLFvRclndmJdMIijLyjFH1Aq2YbbGhEl
# gcL09Wcu53kd9eIcdfROzMf8578LgEcp/8/NabEMC2DrZ+aEG5tN/W1HOsfZwWFh
# 8pUSoQ0HrmMh2PSZHP94VYHupXnoIIJfCtq1UxlUAVcNh5GNwnzxVIaA4WLbgnM+
# Jl7wQBLSOdUmAw2FiDFfCguLAgMBAAGjggF/MIIBezAfBgNVHSUEGDAWBgorBgEE
# AYI3TAgBBggrBgEFBQcDAzAdBgNVHQ4EFgQUpxNdHyGJVegD7p4XNuryVIg1Ga8w
# UQYDVR0RBEowSKRGMEQxDDAKBgNVBAsTA0FPQzE0MDIGA1UEBRMrMjMwMDEyK2M4
# MDRiNWVhLTQ5YjQtNDIzOC04MzYyLWQ4NTFmYTIyNTRmYzAfBgNVHSMEGDAWgBRI
# bmTlUAXTgqoXNzcitW2oynUClTBUBgNVHR8ETTBLMEmgR6BFhkNodHRwOi8vd3d3
# Lm1pY3Jvc29mdC5jb20vcGtpb3BzL2NybC9NaWNDb2RTaWdQQ0EyMDExXzIwMTEt
# MDctMDguY3JsMGEGCCsGAQUFBwEBBFUwUzBRBggrBgEFBQcwAoZFaHR0cDovL3d3
# dy5taWNyb3NvZnQuY29tL3BraW9wcy9jZXJ0cy9NaWNDb2RTaWdQQ0EyMDExXzIw
# MTEtMDctMDguY3J0MAwGA1UdEwEB/wQCMAAwDQYJKoZIhvcNAQELBQADggIBAE2X
# TzR+8XCTnOPVGkucEX5rJsSlJPTfRNQkurNqCImZmssx53Cb/xQdsAc5f+QwOxMi
# 3g7IlWe7bn74fJWkkII3k6aD00kCwaytWe+Rt6dmAA6iTCXU3OddBwLKKDRlOzmD
# rZUqjsqg6Ag6HP4+e0BJlE2OVCUK5bHHCu5xN8abXjb1p0JE+7yHsA3ANdkmh1//
# Z+8odPeKMAQRimfMSzVgaiHnw40Hg16bq51xHykmCRHU9YLT0jYHKa7okm2QfwDJ
# qFvu0ARl+6EOV1PM8piJ858Vk8gGxGNSYQJPV0gc9ft1Esq1+fTCaV+7oZ0NaYMn
# 64M+HWsxw+4O8cSEQ4fuMZwGADJ8tyCKuQgj6lawGNSyvRXsN+1k02sVAiPGijOH
# OtGbtsCWWSygAVOEAV/ye8F6sOzU2FL2X3WBRFkWOCdTu1DzXnHf99dR3DHVGmM1
# Kpd+n2Y3X89VM++yyrwsI6pEHu77Z0i06ELDD4pRWKJGAmEmWhm/XJTpqEBw51sw
# THyA1FBnoqXuDus9tfHleR7h9VgZb7uJbXjiIFgl/+RIs+av8bJABBdGUNQMbJEU
# fe7K4vYm3hs7BGdRLg+kF/dC/z+RiTH4p7yz5TpS3Cozf0pkkWXYZRG222q3tGxS
# /L+LcRbELM5zmqDpXQjBRUWlKYbsATFtXnTGVjELMIIHejCCBWKgAwIBAgIKYQ6Q
# 0gAAAAAAAzANBgkqhkiG9w0BAQsFADCBiDELMAkGA1UEBhMCVVMxEzARBgNVBAgT
# Cldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29m
# dCBDb3Jwb3JhdGlvbjEyMDAGA1UEAxMpTWljcm9zb2Z0IFJvb3QgQ2VydGlmaWNh
# dGUgQXV0aG9yaXR5IDIwMTEwHhcNMTEwNzA4MjA1OTA5WhcNMjYwNzA4MjEwOTA5
# WjB+MQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMH
# UmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSgwJgYDVQQD
# Ex9NaWNyb3NvZnQgQ29kZSBTaWduaW5nIFBDQSAyMDExMIICIjANBgkqhkiG9w0B
# AQEFAAOCAg8AMIICCgKCAgEAq/D6chAcLq3YbqqCEE00uvK2WCGfQhsqa+laUKq4
# BjgaBEm6f8MMHt03a8YS2AvwOMKZBrDIOdUBFDFC04kNeWSHfpRgJGyvnkmc6Whe
# 0t+bU7IKLMOv2akrrnoJr9eWWcpgGgXpZnboMlImEi/nqwhQz7NEt13YxC4Ddato
# 88tt8zpcoRb0RrrgOGSsbmQ1eKagYw8t00CT+OPeBw3VXHmlSSnnDb6gE3e+lD3v
# ++MrWhAfTVYoonpy4BI6t0le2O3tQ5GD2Xuye4Yb2T6xjF3oiU+EGvKhL1nkkDst
# rjNYxbc+/jLTswM9sbKvkjh+0p2ALPVOVpEhNSXDOW5kf1O6nA+tGSOEy/S6A4aN
# 91/w0FK/jJSHvMAhdCVfGCi2zCcoOCWYOUo2z3yxkq4cI6epZuxhH2rhKEmdX4ji
# JV3TIUs+UsS1Vz8kA/DRelsv1SPjcF0PUUZ3s/gA4bysAoJf28AVs70b1FVL5zmh
# D+kjSbwYuER8ReTBw3J64HLnJN+/RpnF78IcV9uDjexNSTCnq47f7Fufr/zdsGbi
# wZeBe+3W7UvnSSmnEyimp31ngOaKYnhfsi+E11ecXL93KCjx7W3DKI8sj0A3T8Hh
# hUSJxAlMxdSlQy90lfdu+HggWCwTXWCVmj5PM4TasIgX3p5O9JawvEagbJjS4NaI
# jAsCAwEAAaOCAe0wggHpMBAGCSsGAQQBgjcVAQQDAgEAMB0GA1UdDgQWBBRIbmTl
# UAXTgqoXNzcitW2oynUClTAZBgkrBgEEAYI3FAIEDB4KAFMAdQBiAEMAQTALBgNV
# HQ8EBAMCAYYwDwYDVR0TAQH/BAUwAwEB/zAfBgNVHSMEGDAWgBRyLToCMZBDuRQF
# TuHqp8cx0SOJNDBaBgNVHR8EUzBRME+gTaBLhklodHRwOi8vY3JsLm1pY3Jvc29m
# dC5jb20vcGtpL2NybC9wcm9kdWN0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNf
# MjIuY3JsMF4GCCsGAQUFBwEBBFIwUDBOBggrBgEFBQcwAoZCaHR0cDovL3d3dy5t
# aWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNSb29DZXJBdXQyMDExXzIwMTFfMDNf
# MjIuY3J0MIGfBgNVHSAEgZcwgZQwgZEGCSsGAQQBgjcuAzCBgzA/BggrBgEFBQcC
# ARYzaHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraW9wcy9kb2NzL3ByaW1hcnlj
# cHMuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwAZQBnAGEAbABfAHAAbwBsAGkAYwB5
# AF8AcwB0AGEAdABlAG0AZQBuAHQALiAdMA0GCSqGSIb3DQEBCwUAA4ICAQBn8oal
# mOBUeRou09h0ZyKbC5YR4WOSmUKWfdJ5DJDBZV8uLD74w3LRbYP+vj/oCso7v0ep
# o/Np22O/IjWll11lhJB9i0ZQVdgMknzSGksc8zxCi1LQsP1r4z4HLimb5j0bpdS1
# HXeUOeLpZMlEPXh6I/MTfaaQdION9MsmAkYqwooQu6SpBQyb7Wj6aC6VoCo/KmtY
# SWMfCWluWpiW5IP0wI/zRive/DvQvTXvbiWu5a8n7dDd8w6vmSiXmE0OPQvyCInW
# H8MyGOLwxS3OW560STkKxgrCxq2u5bLZ2xWIUUVYODJxJxp/sfQn+N4sOiBpmLJZ
# iWhub6e3dMNABQamASooPoI/E01mC8CzTfXhj38cbxV9Rad25UAqZaPDXVJihsMd
# YzaXht/a8/jyFqGaJ+HNpZfQ7l1jQeNbB5yHPgZ3BtEGsXUfFL5hYbXw3MYbBL7f
# QccOKO7eZS/sl/ahXJbYANahRr1Z85elCUtIEJmAH9AAKcWxm6U/RXceNcbSoqKf
# enoi+kiVH6v7RyOA9Z74v2u3S5fi63V4GuzqN5l5GEv/1rMjaHXmr/r8i+sLgOpp
# O6/8MO0ETI7f33VtY5E90Z1WTk+/gFcioXgRMiF670EKsT/7qMykXcGhiJtXcVZO
# SEXAQsmbdlsKgEhr/Xmfwb1tbWrJUnMTDXpQzTGCFdkwghXVAgEBMIGVMH4xCzAJ
# BgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25k
# MR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRpb24xKDAmBgNVBAMTH01pY3Jv
# c29mdCBDb2RlIFNpZ25pbmcgUENBIDIwMTECEzMAAADDDpun2LLc9ywAAAAAAMMw
# DQYJYIZIAWUDBAIBBQCggcgwGQYJKoZIhvcNAQkDMQwGCisGAQQBgjcCAQQwHAYK
# KwYBBAGCNwIBCzEOMAwGCisGAQQBgjcCARUwLwYJKoZIhvcNAQkEMSIEIJjz21aY
# 0ldTzPfOLKYTenxk92JA9QfHWb4PXMooiKleMFwGCisGAQQBgjcCAQwxTjBMoC6A
# LABNAGkAYwByAG8AcwBvAGYAdAAgAEQAeQBuAGEAbQBpAGMAcwAgAE4AQQBWoRqA
# GGh0dHA6Ly93d3cubWljcm9zb2Z0LmNvbTANBgkqhkiG9w0BAQEFAASCAQAfbIUu
# XlTc8Gz6w/GyaVHJPyi9roDFjSgfW+uk7PEkdouMYN9K83FapsbYEz/QK3rsNoos
# ueBZI/6Z+9eoaEIcl6AS27NgqgqZWH4O4n3vSglcLTPXxUhmME2IQoUzbU5+Ux15
# BqdDaS2epM6moAv8OszfzhYFp293D6E37hZjURcIZjRE3U85SbBzErfp4UnCXMLe
# RafKWJU+4eCIVjdAHVuqKCtwkMepPNQsZkulmaoJrRMQRClBkYNsrsqGzCthOHKg
# MwVHG0uGL3wq0Z0/3kHTZ96h54vJsPvjsbCAL9szjBf5+CZMRsTrAN0M7Ma1QHz7
# Z3x8e4Bfp8/R/sR+oYITSTCCE0UGCisGAQQBgjcDAwExghM1MIITMQYJKoZIhvcN
# AQcCoIITIjCCEx4CAQMxDzANBglghkgBZQMEAgEFADCCATwGCyqGSIb3DQEJEAEE
# oIIBKwSCAScwggEjAgEBBgorBgEEAYRZCgMBMDEwDQYJYIZIAWUDBAIBBQAEIBkA
# iqRStFuzItTfWo2OEnSAdH/8Q1y8riT7tEjHA0D/AgZZ1ocDSe4YEzIwMTcxMTIy
# MjA1NDMzLjY2MlowBwIBAYACAfSggbikgbUwgbIxCzAJBgNVBAYTAlVTMRMwEQYD
# VQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNy
# b3NvZnQgQ29ycG9yYXRpb24xDDAKBgNVBAsTA0FPQzEnMCUGA1UECxMebkNpcGhl
# ciBEU0UgRVNOOjU3QzgtMkQxNS0xQzhCMSUwIwYDVQQDExxNaWNyb3NvZnQgVGlt
# ZS1TdGFtcCBTZXJ2aWNloIIOzTCCBnEwggRZoAMCAQICCmEJgSoAAAAAAAIwDQYJ
# KoZIhvcNAQELBQAwgYgxCzAJBgNVBAYTAlVTMRMwEQYDVQQIEwpXYXNoaW5ndG9u
# MRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVNaWNyb3NvZnQgQ29ycG9yYXRp
# b24xMjAwBgNVBAMTKU1pY3Jvc29mdCBSb290IENlcnRpZmljYXRlIEF1dGhvcml0
# eSAyMDEwMB4XDTEwMDcwMTIxMzY1NVoXDTI1MDcwMTIxNDY1NVowfDELMAkGA1UE
# BhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQxHjAc
# BgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9zb2Z0
# IFRpbWUtU3RhbXAgUENBIDIwMTAwggEiMA0GCSqGSIb3DQEBAQUAA4IBDwAwggEK
# AoIBAQCpHQ28dxGKOiDs/BOX9fp/aZRrdFQQ1aUKAIKF++18aEssX8XD5WHCdrc+
# Zitb8BVTJwQxH0EbGpUdzgkTjnxhMFmxMEQP8WCIhFRDDNdNuDgIs0Ldk6zWczBX
# JoKjRQ3Q6vVHgc2/JGAyWGBG8lhHhjKEHnRhZ5FfgVSxz5NMksHEpl3RYRNuKMYa
# +YaAu99h/EbBJx0kZxJyGiGKr0tkiVBisV39dx898Fd1rL2KQk1AUdEPnAY+Z3/1
# ZsADlkR+79BL/W7lmsqxqPJ6Kgox8NpOBpG2iAg16HgcsOmZzTznL0S6p/TcZL2k
# AcEgCZN4zfy8wMlEXV4WnAEFTyJNAgMBAAGjggHmMIIB4jAQBgkrBgEEAYI3FQEE
# AwIBADAdBgNVHQ4EFgQU1WM6XIoxkPNDe3xGG8UzaFqFbVUwGQYJKwYBBAGCNxQC
# BAweCgBTAHUAYgBDAEEwCwYDVR0PBAQDAgGGMA8GA1UdEwEB/wQFMAMBAf8wHwYD
# VR0jBBgwFoAU1fZWy4/oolxiaNE9lJBb186aGMQwVgYDVR0fBE8wTTBLoEmgR4ZF
# aHR0cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvTWljUm9v
# Q2VyQXV0XzIwMTAtMDYtMjMuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggrBgEFBQcw
# AoY+aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNSb29DZXJB
# dXRfMjAxMC0wNi0yMy5jcnQwgaAGA1UdIAEB/wSBlTCBkjCBjwYJKwYBBAGCNy4D
# MIGBMD0GCCsGAQUFBwIBFjFodHRwOi8vd3d3Lm1pY3Jvc29mdC5jb20vUEtJL2Rv
# Y3MvQ1BTL2RlZmF1bHQuaHRtMEAGCCsGAQUFBwICMDQeMiAdAEwAZQBnAGEAbABf
# AFAAbwBsAGkAYwB5AF8AUwB0AGEAdABlAG0AZQBuAHQALiAdMA0GCSqGSIb3DQEB
# CwUAA4ICAQAH5ohRDeLG4Jg/gXEDPZ2joSFvs+umzPUxvs8F4qn++ldtGTCzwsVm
# yWrf9efweL3HqJ4l4/m87WtUVwgrUYJEEvu5U4zM9GASinbMQEBBm9xcF/9c+V4X
# NZgkVkt070IQyK+/f8Z/8jd9Wj8c8pl5SpFSAK84Dxf1L3mBZdmptWvkx872ynoA
# b0swRCQiPM/tA6WWj1kpvLb9BOFwnzJKJ/1Vry/+tuWOM7tiX5rbV0Dp8c6ZZpCM
# /2pif93FSguRJuI57BlKcWOdeyFtw5yjojz6f32WapB4pm3S4Zz5Hfw42JT0xqUK
# loakvZ4argRCg7i1gJsiOCC1JeVk7Pf0v35jWSUPei45V3aicaoGig+JFrphpxHL
# mtgOR5qAxdDNp9DvfYPw4TtxCd9ddJgiCGHasFAeb73x4QDf5zEHpJM692VHeOj4
# qEir995yfmFrb3epgcunCaw5u+zGy9iCtHLNHfS4hQEegPsbiSpUObJb2sgNVZl6
# h3M7COaYLeqN4DMuEin1wC9UJyH3yKxO2ii4sanblrKnQqLJzxlBTeCG+SqaoxFm
# MNO7dDJL32N79ZmKLxvHIa9Zta7cRDyXUHHXodLFVeNp3lfB0d4wwP3M5k37Db9d
# T+mdHhk4L7zPWAUu7w2gUDXa7wknHNWzfjUeCLraNtvTX4/edIhJEjCCBNkwggPB
# oAMCAQICEzMAAACqt6mI/+pXwwoAAAAAAKowDQYJKoZIhvcNAQELBQAwfDELMAkG
# A1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1JlZG1vbmQx
# HjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEmMCQGA1UEAxMdTWljcm9z
# b2Z0IFRpbWUtU3RhbXAgUENBIDIwMTAwHhcNMTYwOTA3MTc1NjUzWhcNMTgwOTA3
# MTc1NjUzWjCBsjELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAO
# BgNVBAcTB1JlZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEM
# MAoGA1UECxMDQU9DMScwJQYDVQQLEx5uQ2lwaGVyIERTRSBFU046NTdDOC0yRDE1
# LTFDOEIxJTAjBgNVBAMTHE1pY3Jvc29mdCBUaW1lLVN0YW1wIFNlcnZpY2UwggEi
# MA0GCSqGSIb3DQEBAQUAA4IBDwAwggEKAoIBAQCe2H97V3j6BfqjmreQr5q51bZ0
# dpHLuM67Ks4rKe9ftGt1mHUogK2Yj7mq8J3StptHZJjGpVmqkXj5Hot59yR0Ruok
# /mkErr7rf1NTE1nVD/z7zd32B0GbGzc32Vx2md2ux8Pb061owlCcdA5Edy4unW/B
# He8sxzt0rtuXLY+BxUcFQMBKN0iTvAOaS/CNBMtDcPvoWLGdrFDnubWE8cebB+z3
# mmnbr0h/rFkxhmIWcJVe8cvMOkfz6j8CjBC7vSOMqQNzTw64DFzxL6p/+mSbjC6Y
# FMBGzyZPPAYjxk8it5IPGgIZ/JwpzYTega/w/a2YrKWIg4aVVIa0m3FAxi83AgMB
# AAGjggEbMIIBFzAdBgNVHQ4EFgQUCLhyHRUk9ZMC4Z/SHEzTq8xZybAwHwYDVR0j
# BBgwFoAU1WM6XIoxkPNDe3xGG8UzaFqFbVUwVgYDVR0fBE8wTTBLoEmgR4ZFaHR0
# cDovL2NybC5taWNyb3NvZnQuY29tL3BraS9jcmwvcHJvZHVjdHMvTWljVGltU3Rh
# UENBXzIwMTAtMDctMDEuY3JsMFoGCCsGAQUFBwEBBE4wTDBKBggrBgEFBQcwAoY+
# aHR0cDovL3d3dy5taWNyb3NvZnQuY29tL3BraS9jZXJ0cy9NaWNUaW1TdGFQQ0Ff
# MjAxMC0wNy0wMS5jcnQwDAYDVR0TAQH/BAIwADATBgNVHSUEDDAKBggrBgEFBQcD
# CDANBgkqhkiG9w0BAQsFAAOCAQEAYI8O7gfOGuF9n3nGeA6dzkykAZ1qZ0eXERuK
# csrkwXznLeOPkWe86HR6P8rpRiQAP6HO8H5P0vQaffR2OB0UNh2l2YlvysVTFO8T
# LCACldEKecXS7m08P5FG6blS3t9c4pykTVFqHcLpk01GchYm+YT/k3fd6AM9VPzC
# BKBfj4e9VXSa6WSssOQaylw7IB8LVVgIsMPLp7xZLE1Cke1bszAukqeTjk6ADK6p
# eTHsUpF8lRCvf8HOI9sPmcxqw8T0LB91ZIIsoNgOB/eaDmWoXJWBnH5Y7nnzkSGt
# 280sv7WIcv4GG51fdg92MoiuUjxtOS7MBk4kSS0vqWYA1vxoJqGCA3cwggJfAgEB
# MIHioYG4pIG1MIGyMQswCQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQ
# MA4GA1UEBxMHUmVkbW9uZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9u
# MQwwCgYDVQQLEwNBT0MxJzAlBgNVBAsTHm5DaXBoZXIgRFNFIEVTTjo1N0M4LTJE
# MTUtMUM4QjElMCMGA1UEAxMcTWljcm9zb2Z0IFRpbWUtU3RhbXAgU2VydmljZaIl
# CgEBMAkGBSsOAwIaBQADFQCcnMVrmMynwa1HIOzTRPzObdDqA6CBwTCBvqSBuzCB
# uDELMAkGA1UEBhMCVVMxEzARBgNVBAgTCldhc2hpbmd0b24xEDAOBgNVBAcTB1Jl
# ZG1vbmQxHjAcBgNVBAoTFU1pY3Jvc29mdCBDb3Jwb3JhdGlvbjEMMAoGA1UECxMD
# QU9DMScwJQYDVQQLEx5uQ2lwaGVyIE5UUyBFU046MjY2NS00QzNGLUM1REUxKzAp
# BgNVBAMTIk1pY3Jvc29mdCBUaW1lIFNvdXJjZSBNYXN0ZXIgQ2xvY2swDQYJKoZI
# hvcNAQEFBQACBQDdv8E0MCIYDzIwMTcxMTIyMDkyNjEyWhgPMjAxNzExMjMwOTI2
# MTJaMHcwPQYKKwYBBAGEWQoEATEvMC0wCgIFAN2/wTQCAQAwCgIBAAICMX0CAf8w
# BwIBAAICGgQwCgIFAN3BErQCAQAwNgYKKwYBBAGEWQoEAjEoMCYwDAYKKwYBBAGE
# WQoDAaAKMAgCAQACAxbjYKEKMAgCAQACAx6EgDANBgkqhkiG9w0BAQUFAAOCAQEA
# YlkObz6MKwDFRkhiHUa3MHoejMgy7nZ724iFq7UjudDA3Yuy5SwQqqOCMbB1onZp
# uIuyfXrbAjyFXXFKFwnWgUeZCjhOQ5ZvqMqczx+TLRtfekUK8WMUqt5zQ8sWGmxu
# fZvGh0FCNiBmcMXx6qviE1IQ3gyuPfkyrn53T8F6J0Puh3FwLZZwFDVCZU3olNeJ
# WJxXj9RedaCRVptPNBuLkOD8Xa3gMsl+GPQ8cwut4sa8YoOe1k2uPCkbGFg3Nk3P
# +F2V6qXmzrxZedCuRfmGJ27avJuXklmSFQpzRke7y9326g0MrDPbECwTAwqh9vBQ
# h604lUwub7gdXO6aWiGC1DGCAvUwggLxAgEBMIGTMHwxCzAJBgNVBAYTAlVTMRMw
# EQYDVQQIEwpXYXNoaW5ndG9uMRAwDgYDVQQHEwdSZWRtb25kMR4wHAYDVQQKExVN
# aWNyb3NvZnQgQ29ycG9yYXRpb24xJjAkBgNVBAMTHU1pY3Jvc29mdCBUaW1lLVN0
# YW1wIFBDQSAyMDEwAhMzAAAAqrepiP/qV8MKAAAAAACqMA0GCWCGSAFlAwQCAQUA
# oIIBMjAaBgkqhkiG9w0BCQMxDQYLKoZIhvcNAQkQAQQwLwYJKoZIhvcNAQkEMSIE
# IHjxgwIrmDxx5uJ4q7HPkRgobtbeba58odkGqhqbCCr4MIHiBgsqhkiG9w0BCRAC
# DDGB0jCBzzCBzDCBsQQUnJzFa5jMp8GtRyDs00T8zm3Q6gMwgZgwgYCkfjB8MQsw
# CQYDVQQGEwJVUzETMBEGA1UECBMKV2FzaGluZ3RvbjEQMA4GA1UEBxMHUmVkbW9u
# ZDEeMBwGA1UEChMVTWljcm9zb2Z0IENvcnBvcmF0aW9uMSYwJAYDVQQDEx1NaWNy
# b3NvZnQgVGltZS1TdGFtcCBQQ0EgMjAxMAITMwAAAKq3qYj/6lfDCgAAAAAAqjAW
# BBS6JQbZZHKkJ9SDrnbbKiMsn1AU2zANBgkqhkiG9w0BAQsFAASCAQAsCVfq/Qg6
# uE7qtnSV5lrOD6ttpnIBgzsStmXyXhHLwBl11Y/wW4P1N1yj0yV616U2ZNbxSRoY
# szDA3NL3aMro1OyeldZtwyFpwo5melZumBaTkx0MnuWOdIrk3yOw8zarrlvxfy32
# kcK64S8pMhbaPJUFyG2OqTXJxemhPkzAggWPkgeRRoDgiWjst2fUZ35R1RpSzVXQ
# xOzjxD+smNAdkh84c6U7TVH9BLDko/oX34t2YPHZDBuUjOq23JQMQcdP9j6HU1Qg
# E35MwHjMQI36Q1vOLprB2jNDX268z/XADhL4a0IYZ1puNGB/kHeSsSaIY21UzWGr
# 3pdfaFJKw3AR
# SIG # End signature block
